function [F,Fexp,copt,Mmopt,Plyapopt] = ...
    robfiltLPVsyn_lmilab_matElim(P,PSIICT,nw,nv,ny,varargin)
%%ROBFILTSYN synthesizes a robust filter.
% This function uses a matrix elimination to solve for the filter.
fprintf('\n - Synthesizing in LMILab matrix elimination mode.\n');
%########### RELEASE VERSION ###########
% function [F,gm,Mm,Plyap,cvx_status] = robfiltsyn(PICT,PSIICT)
% Input arguments:
% (1) PICT is the interconnection of the nominal plant and psi filters.
% (2) PSIICT is the interconnection of the PSI filters.
% Output arguments:
% (1) F is the synthesized filter.
% (2) copt is the L2-L2 norm of lft(DELTA,G).
% (3) Mmopt is the IQC multiplier in the time domain (single matrix).
% (4) Plyapopt is the Lyapunov function matrix for the plant-filter ICT.

%% Manage variable input arguments
if ~isempty(varargin)
    if strcmpi(varargin{1},'ff')
        gn = -1; % feedforward problem
    end
    MULT = varargin{2};
else
    gn = +1; % estimator problem
    MULT = 'realp'; % constant real parametric uncertainty
end

%% Interconnection of plant and IQC filters
PICT = connect(P,PSIICT,P.InputName,...
    [PSIICT.OutputName;P.OutputName(nv+1:end)]);
[nr,nc] = size(PICT);
% nr and nc are used in next section because 'end' does not work with PSS

%% Extract SS matrices
[A,B,C,D] = ssdata(PICT);

n = size(A,1); % state dimension
B1 = B(:,1:nw); % uncertain inputs
B2 = B(:,nw+1:nc); nd = size(B2,2); % disturbance inputs
nzPsi = size(PSIICT,1); % output size of IQC filters
C1 = C(1:nzPsi,:); % filtered uncertain outputs
C2 = C(nzPsi+1:nzPsi+ny,:); % measurements sent to estimator
p = ny; % size of measurements sent to estimator
C3 = gn.*C(nzPsi+ny+1:nr,:); % outputs to be estimated
nq = size(C3,1); % size of outputs to be estimated
D11 = D(1:nzPsi,1:nw);
D12 = D(1:nzPsi,nw+1:nc);
D21 = D(nzPsi+1:nzPsi+ny,1:nw);
D22 = D(nzPsi+1:nzPsi+ny,nw+1:nc);
D31 = gn.*D(nzPsi+ny+1:nr,1:nw);
D32 = gn.*D(nzPsi+ny+1:nr,nw+1:nc);

% A = PICT.a; n = size(A,1); % state dimension
% B1 = PICT(:,1:nw).b; % uncertain inputs
% B2 = PICT(:,nw+1:nc).b; nd = size(B2,2); % disturbance inputs
% nzPsi = size(PSIICT,1); % output size of IQC filters
% C1 = PICT(1:nzPsi,:).c; % filtered uncertain outputs
% C2 = PICT(nzPsi+1:nzPsi+ny,:).c; % measurements sent to estimator
% p = ny; % size of measurements sent to estimator
% C3 = gn.*(PICT(nzPsi+ny+1:nr,:).c); % outputs to be estimated
% nq = size(C3,1); % size of outputs to be estimated
% D11 = PICT(1:nzPsi,1:nw).d;
% D12 = PICT(1:nzPsi,nw+1:nc).d;
% D21 = PICT(nzPsi+1:nzPsi+ny,1:nw).d;
% D22 = PICT(nzPsi+1:nzPsi+ny,nw+1:nc).d;
% D31 = gn.*(PICT(nzPsi+ny+1:nr,1:nw).d);
% D32 = gn.*(PICT(nzPsi+ny+1:nr,nw+1:nc).d);
% [AR,BR,CR,DR] = ssdata(R); nstate_R = size(AR,1); ninput_R = size(DR,2);
% nPt = size(Pt.a,1); % state dimension of Pt
% Number of grid points in PSS object (currently only supports 1D grid)
ng = size(PICT.Parameter.rho.GridData,1); % number of grid points

%% Extracting (1,1) block from PSIICT
switch MULT
    case 'realp'
        PSI11 = ss(PSIICT('Uncertainv','v')); % extract the (1,1) block of PSIICT
        [AR,BR,CR,DR] = ssdata(PSI11); % extract ss data of PSI11
        nstate_R = size(AR,1); % number of states in PSI11
        ninput_R = size(DR,2); % number of inputs in PSI11
end

%% LMILab implementation
% Set Uflag = 1 to enforce U>=0 but not KYP_R>=0
% Set Uflag = 0 to enforce KYP_R>=0 but not U>=0
Uflag = 1;

% Initialize
setlmis([]);

% Definition of variables in LMI
X = lmivar(1,[n,1]); % (n,n) symmetric
Z = lmivar(1,[n,1]); % (n,n) symmetric
switch MULT
    case 'realp'
        [Pm,ndec,Pmdec] = lmivar(1,[nzPsi/2,1]); % (nzPsi/2,nzPsi/2) symmetric
        if nzPsi > 2
            [Rm,~,Rmdec] = lmivar(3,skewdec(nzPsi/2,ndec)); % (nzPsi/2,nzPsi/2) skew-symmetric
        else
            Rmdec = 0;
        end
        Mm = lmivar(3,[Pmdec,Rmdec;-Rmdec,-Pmdec]); % as given in Scherer_08Automatica
    case {'zamesfalb','sector'}
        [lm,~,lmdec] = lmivar(1,[1 1]);
        Mm = lmivar(3,lmdec*blkdiag(eye(nv),-eye(nw)));
end

[gam,ngam,gamdec] = lmivar(1,[1,1]); % (1,1) symmetric
gam_Ind = lmivar(3,gamdec*eye(nd));
gam_Inq = lmivar(3,gamdec*eye(nq));

% Define LMIs at each grid point
NQHNQ = zeros(ng,1); % ng grid points
NRHNR = zeros(ng,1); % ng grid points

for kk = 1:ng
    % Extract data at a grid point
    Ap = A.Data(:,:,kk);
    B1p = B1.Data(:,:,kk);
    B2p = B2.Data(:,:,kk);
    C1p = C1.Data(:,:,kk);
    C2p = C2.Data(:,:,kk);
    C3p = C3.Data(:,:,kk);
    D11p = D11.Data(:,:,kk);
    D12p = D12.Data(:,:,kk);
    D21p = D21.Data(:,:,kk);
    D22p = D22.Data(:,:,kk);
    D31p = D31.Data(:,:,kk);
    D32p = D32.Data(:,:,kk);
    
    % Define LMI #1: NQ'*H*NQ < 0
    NQHNQ(kk,1) = newlmi;
    lmiterm([+NQHNQ(kk,1) 1 1 Z],Ap',eye(n),'s'); % A'*Z + Z*A
    lmiterm([+NQHNQ(kk,1) 1 2 Z],eye(n),B1p); % Z*B1
    lmiterm([+NQHNQ(kk,1) 1 3 Z],eye(n),B2p); % Z*B2
    lmiterm([-NQHNQ(kk,1) 3 3 gam_Ind],eye(nd),eye(nd)); % -gm.*eye(nd)
    
    lmiterm([+NQHNQ(kk,1) 1 1 Mm],C1p',C1p); % C1'*Mm*C1
    lmiterm([+NQHNQ(kk,1) 1 2 Mm],C1p',D11p); % C1'*Mm*D11
    lmiterm([+NQHNQ(kk,1) 1 3 Mm],C1p',D12p); % C1'*Mm*D12
    lmiterm([+NQHNQ(kk,1) 2 2 Mm],D11p',D11p); % D11'*Mm*D11
    lmiterm([+NQHNQ(kk,1) 2 3 Mm],D11p',D12p); % D11'*Mm*D12
    lmiterm([+NQHNQ(kk,1) 3 3 Mm],D12p',D12p); % D12'*Mm*D12
    
    % Define LMI #2: NR'*H*NR < 0
    NRHNR(kk,1) = newlmi;
    
%     N = null([C2p, D21p, D22p]); % numerically computed
    N = [[zeros(n,nw); eye(nw); -pinv(D22p)*D21p],...
        [zeros(n+nw,nd-ny); null(D22p)],...
        [eye(n); zeros(nw,n); -pinv(D22p)*C2p]]; % explicit expression
    NOF = blkdiag(N,eye(nq));
    
    lmiterm([+NRHNR(kk,1) 0 0 0],NOF); % outer factor NOF
    lmiterm([+NRHNR(kk,1) 1 1 X],Ap',eye(n),'s'); % A'*X + X*A
    lmiterm([+NRHNR(kk,1) 1 2 X],eye(n),B1p); % X*B1
    lmiterm([+NRHNR(kk,1) 1 3 X],eye(n),B2p); % X*B2
    lmiterm([+NRHNR(kk,1) 1 4 0],-C3p'); % -C3'
    lmiterm([+NRHNR(kk,1) 2 4 0],-D31p'); % -D31'
    lmiterm([+NRHNR(kk,1) 3 3 gam_Ind],-eye(nd),eye(nd)); % -gm.*eye(nd)
    lmiterm([+NRHNR(kk,1) 3 4 0],-D32p'); % -D32'
    lmiterm([+NRHNR(kk,1) 4 4 gam_Inq],-eye(nq),eye(nq)); % -gm.*eye(nq)
    
    lmiterm([+NRHNR(kk,1) 1 1 Mm],C1p',C1p); % C1'*Mm*C1
    lmiterm([+NRHNR(kk,1) 1 2 Mm],C1p',D11p); % C1'*Mm*D11
    lmiterm([+NRHNR(kk,1) 1 3 Mm],C1p',D12p); % C1'*Mm*D12
    lmiterm([+NRHNR(kk,1) 2 2 Mm],D11p',D11p); % D11'*Mm*D11
    lmiterm([+NRHNR(kk,1) 2 3 Mm],D11p',D12p); % D11'*Mm*D12
    lmiterm([+NRHNR(kk,1) 3 3 Mm],D12p',D12p); % D12'*Mm*D12
end

% Define LMI #2: 0 < KYP_R
KYP_R = newlmi;
switch MULT
    case 'realp'
        if nstate_R>0 && ~Uflag
            lmiterm([-KYP_R 1 1 Y],AR',eye(nstate_R),'s'); % AR'*Y + Y*AR
            lmiterm([-KYP_R 1 2 Y],eye(nstate_R),BR); % Y*BR
            
            lmiterm([-KYP_R 1 1 Pm],CR',CR); % CR'*Pm*CR
            lmiterm([-KYP_R 1 2 Pm],CR',DR); % CR'*Pm*DR
            lmiterm([-KYP_R 2 2 Pm],DR',DR); % DR'*Pm*DR
        else
            lmiterm([-KYP_R 1 1 Pm],1,1); % Pm
        end
    case {'zamesfalb','sector'}
        lmiterm([-KYP_R 1 1 lm],1,1); % enforce lm>=0
end

% Define LMI #3: Z < X (X3 > 0 ensures a stable filter is synthesized)
LMI_X3 = newlmi;
lmiterm([+LMI_X3 1 1 Z],1,1); % Z
lmiterm([-LMI_X3 1 1 X],1,1); % X

% Gain minimization subject to LMI constraints: mincx solver
lmisys = getlmis;
ndec = decnbr(lmisys); % total number of decision variables
c = zeros(ndec,1); % weighting vector for decision variables
c(ngam) = 1; % pick gamma as the cost function
% lmiopt = [];             % Default options (Display on)
lmiopt = [0 0 0 0 1];   % Turn off LMILab Display
% lmiopt = [1e-4 400 100 100 1];
% lmiopt = [1e-4 250 -1 50 0]; % Custom options (Display on)
[copt,xopt] = mincx(lmisys,c,lmiopt);

%% Extract optimal values of decision variables
Xopt = dec2mat(lmisys,xopt,X);
Zopt = dec2mat(lmisys,xopt,Z);
switch MULT
    case 'realp'
        Pmopt = dec2mat(lmisys,xopt,Pm);
        if exist('Rm','var')
            Rmopt = dec2mat(lmisys,xopt,Rm);
        else
            Rmopt = 0;
        end
        Mmopt = [Pmopt,Rmopt;Rmopt',-Pmopt];
    case {'zamesfalb','sector'}
        lmopt = dec2mat(lmisys,xopt,lm);
        Mmopt = lmopt*blkdiag(eye(nv),-eye(nw));
end
gmopt = dec2mat(lmisys,xopt,gam);
switch MULT
    case {'realp'}
        if nstate_R > 0 && ~Uflag
            Yopt = dec2mat(lmisys,xopt,Y);
        end
end

%% Lyapunov matrix reconstruction (implicit)
% SCHERER: LRM = 1 for X2 = eye(n); X3 = inv(X - Z);
% GAHINET: LRM = 0 for X2 = chol(X - Z, 'lower'); X3 = eye(n);
LRM = 0;

if LRM
    X2opt = eye(n);
    X3opt = inv(Xopt - Zopt);
else
    X2opt = chol(Xopt - Zopt,'lower');
    X3opt = eye(n);
end
Plyapopt = [Xopt, X2opt; X2opt', X3opt];

%% Filter reconstruction (implicit, using basiclmi)
% Filter state-space matrices at each grid point
Kcell = cell(ng,1);
Fcell = cell(ng,1);

for kk = 1:ng
    % Extract data at a grid point
    Ap = A.Data(:,:,kk);
    B1p = B1.Data(:,:,kk);
    B2p = B2.Data(:,:,kk);
    C1p = C1.Data(:,:,kk);
    C2p = C2.Data(:,:,kk);
    C3p = C3.Data(:,:,kk);
    D11p = D11.Data(:,:,kk);
    D12p = D12.Data(:,:,kk);
    D21p = D21.Data(:,:,kk);
    D22p = D22.Data(:,:,kk);
    D31p = D31.Data(:,:,kk);
    D32p = D32.Data(:,:,kk);
    
    % Reconstruct matrices that existed before elimination
    H = [Ap'*Zopt + Zopt*Ap, Ap'*Xopt + Zopt*Ap, Zopt*B1p, Zopt*B2p, -C3p';...
        Ap'*Zopt + Xopt*Ap, Ap'*Xopt + Xopt*Ap, Xopt*B1p, Xopt*B2p, -C3p';...
        B1p'*Zopt, B1p'*Xopt, zeros(nw,nw), zeros(nw,nd), -D31p';...
        B2p'*Zopt, B2p'*Xopt, zeros(nd,nw), -gmopt.*eye(nd), -D32p';...
        -C3p, -C3p, -D31p, -D32p, -gmopt.*eye(nq)]...
        +[C1p, C1p, D11p, D12p, zeros(size(C1p,1),nq)]'*Mmopt...
        *[C1p, C1p, D11p, D12p, zeros(size(C1p,1),nq)];
    Q = [zeros(n,n), eye(n), zeros(n,nw+nd+nq);...
        zeros(nq,2*n+nw+nd), eye(nq)];
    R = [eye(n), zeros(n,n+nw+nd+nq);...
        C2p, C2p, D21p, D22p, zeros(p,nq)];
    
    % Call basiclmi to extract transformed controller matrices
    Kcell{kk,1} = basiclmi(H,Q,R,'Xmin');
    A1bar = Kcell{kk,1}(1:n,1:n); % transformed A matrix
    Bbar = Kcell{kk,1}(1:n,n+1:end); % transformed B matrix
    Cbar = Kcell{kk,1}(n+1:end,1:n); % transformed C matrix
    Dbar = Kcell{kk,1}(n+1:end,n+1:end); % transformed D matrix
    
    % Filter state-space matrices
    DF = Dbar;
    CF = -Cbar*inv(X2opt')*X3opt;
    BF = inv(X2opt)*Bbar;
    AF = -inv(X2opt)*A1bar*inv(X2opt')*X3opt;
    Fcell{kk,1} = ss(AF,BF,CF,DF);
end

Farray = stack(1,Fcell{:}); % create SS array
F = pss(Farray,PICT.Domain); % create LPV filter

%% Input/output grouping for filter
% state names of filter
FStateName = cell(1,n);
for k = 1:n
    FStateName{1,k} = ['xf(',num2str(k),')'];
end

% input (y) to filter
if length(F.InputName) == 1
    FInpName = 'y(1)';
else
    FInpName = 'y'; % automatic vectorization expected
end

% output (xhat) from filter
if length(F.OutputName) == 1
    FOutName = 'zhat(1)';
else
    FOutName = 'zhat'; % automatic vectorization expected
end

% Set properties for F
F = set(F,'StateName',FStateName,...
    'InputName',FInpName,'OutputName',FOutName,...
    'InputGroup',struct('Measurement',1:length(F.InputName)),...
    'OutputGroup',struct('OutputEstimate',1:length(F.OutputName)));

%% Filter reconstruction (explicit)
% Lyapunov matrix reconstruction (explicit)
Wopt = Xopt - Zopt;
% Pexp = [Xopt, -Wopt; -Wopt, Wopt];

% Filter state-space matrices at each grid point
Fexpcell = cell(ng,1);

for kk = 1:ng
    % Extract data at a grid point
    Ap = A.Data(:,:,kk);
    B1p = B1.Data(:,:,kk);
    B2p = B2.Data(:,:,kk);
    C1p = C1.Data(:,:,kk);
    C2p = C2.Data(:,:,kk);
    C3p = C3.Data(:,:,kk);
    D11p = D11.Data(:,:,kk);
    D12p = D12.Data(:,:,kk);
    D21p = D21.Data(:,:,kk);
    D22p = D22.Data(:,:,kk);
    D31p = D31.Data(:,:,kk);
    D32p = D32.Data(:,:,kk);
    
    % Filter reconstruction by Gahinet and Apkarian
    Lopt = [Ap'*Xopt + Xopt*Ap, -Ap'*Wopt, Xopt*B1p, Xopt*B2p, -C3p';...
        -Wopt*Ap, zeros(n,n), -Wopt*B1p, -Wopt*B2p, zeros(n,nq);...
        B1p'*Xopt, -B1p'*Wopt, zeros(nw,nw), zeros(nw,nd), -D31p';...
        B2p'*Xopt, -B2p'*Wopt, zeros(nd,nw), -gmopt.*eye(nd), -D32p';...
        -C3p, zeros(nq,n), -D31p, -D32p, -gmopt.*eye(nq)] + ...
        [C1p, zeros(nzPsi,n), D11p, D12p, zeros(nzPsi,nq)]'*...
        Mmopt*...
        [C1p, zeros(nzPsi,n), D11p, D12p, zeros(nzPsi,nq)];
    
    URQ = [zeros(2*n,nw+nd-ny);...
        eye(nw), zeros(nw,nd-ny);...
        -pinv(D22p)*D21p, null(D22p);...
        zeros(nq,nw+nd-ny)];
    
    UR = [eye(n), zeros(n,nq);...
        zeros(n+nw,n+nq);...
        -pinv(D22p)*C2p, zeros(nd,nq);...
        zeros(nq,n), eye(nq)];
    
    UQ = [inv(Zopt), zeros(n,ny);...
        inv(Zopt), zeros(n,ny);...
        zeros(nw,n+ny);...
        zeros(nd,n), D22p';...
        zeros(nq,n+ny)];
    
    R1 = [inv(Zopt), zeros(n,ny);...
        C2p/Zopt, D22p*D22p'];
    
    Q1 = blkdiag(-Wopt,eye(nq));
    
    L11 = URQ'*Lopt*URQ;
    L12 = URQ'*Lopt*UR;
    L13 = URQ'*Lopt*UQ;
    L23 = UR'*Lopt*UQ;
    
    THT_1 = -L23+L12'*(L11\L13);
    THT = (Q1'\THT_1)/R1;
    
    % Filter state-space matrices (numeric)
    AF = THT(1:n,1:n);
    BF = THT(1:n,n+1:end);
    CF = THT(n+1:end,1:n);
    DF = THT(n+1:end,n+1:end);
    
    % Filter state-space matrices (analytic)
%     DF = zeros(nq,ny);
%     CF = C3p - DF*C2p;
%     BF = Wopt\(C2p');
%     AF = Ap - BF*C2p;
    
    Fexpcell{kk,1} = ss(AF,BF,CF,DF);
end

% Create LPV filter
Fexparray = stack(1,Fexpcell{:}); % create SS array
Fexp = pss(Fexparray,PICT.Domain); % create LPV filter

% Set properties for F
Fexp = set(Fexp,'StateName',FStateName,...
    'InputName',FInpName,'OutputName',FOutName,...
    'InputGroup',struct('Measurement',1:length(Fexp.InputName)),...
    'OutputGroup',struct('OutputEstimate',1:length(Fexp.OutputName)));

%% Verify LMI feasibility through reconstruction
% Reconstruct LMIs at each grid point
KYP_LMI_opt = cell(ng,2); % ng grid points, second column stores eigenvalue
for kk = 1:ng
    % Extract data at a grid point
    Ap = A.Data(:,:,kk);
    B1p = B1.Data(:,:,kk);
    B2p = B2.Data(:,:,kk);
    C1p = C1.Data(:,:,kk);
    C2p = C2.Data(:,:,kk);
    C3p = C3.Data(:,:,kk);
    D11p = D11.Data(:,:,kk);
    D12p = D12.Data(:,:,kk);
    D21p = D21.Data(:,:,kk);
    D22p = D22.Data(:,:,kk);
    D31p = D31.Data(:,:,kk);
    D32p = D32.Data(:,:,kk);
    
    % Extract transformed filter matrices at a grid point
    A1bar = Kcell{kk,1}(1:n,1:n); % transformed A matrix
    Bbar = Kcell{kk,1}(1:n,n+1:end); % transformed B matrix
    Cbar = Kcell{kk,1}(n+1:end,1:n); % transformed C matrix
    Dbar = Kcell{kk,1}(n+1:end,n+1:end); % transformed D matrix
    Abar = A1bar + Bbar*C2p;
    
    % Reconstruct KYP_LMI
    THp1T = ...
        [Ap'*Zopt + Zopt*Ap, Ap'*Xopt + Zopt*Ap, Zopt*B1p, Zopt*B2p, -C3p';...
        Ap'*Zopt + Xopt*Ap, Ap'*Xopt + Xopt*Ap, Xopt*B1p, Xopt*B2p, -C3p';...
        B1p'*Zopt, B1p'*Xopt, zeros(nw,nw), zeros(nw,nd), -D31p';...
        B2p'*Zopt, B2p'*Xopt, zeros(nd,nw), -gmopt.*eye(nd), -D32p';...
        -C3p, -C3p, -D31p, -D32p, -gmopt.*eye(nq)];
    
    THp2T = [([C1p, C1p, D11p, D12p]'*...
        Mmopt*...
        [C1p, C1p, D11p, D12p]),...
        zeros(2*n+nw+nd,nq);...
        zeros(nq,2*n+nw+nd), zeros(nq,nq)];
    
    TQFRT = ...
        [zeros(n,2*n+nw+nd+nq);...
        Abar, Bbar*C2p, Bbar*D21p, Bbar*D22p, zeros(n,nq);...
        zeros(nw,2*n+nw+nd+nq);...
        zeros(nd,2*n+nw+nd+nq);...
        Dbar*C2p + Cbar, Dbar*C2p, Dbar*D21p, Dbar*D22p, zeros(nq,nq)];
    
    TRFQT = TQFRT';
    
    KYP_LMI_opt{kk,1} = THp1T + THp2T + TQFRT + TRFQT;
    KYP_LMI_opt{kk,2} = -max(real(eig(KYP_LMI_opt{kk,1})));
end

% Reconstruct KYP_R
switch MULT
    case 'realp'
        if nstate_R > 0 && ~Uflag
            KYP_R_opt = [AR'*Yopt + Yopt*AR,Yopt*BR;...
                BR'*Yopt,zeros(ninput_R)] + ...
                [CR,DR]'*Pmopt*[CR,DR];
        else
            KYP_R_opt = Pmopt;
        end
    case {'zamesfalb','sector'}
        KYP_R_opt = lmopt;
end

% Reconstruct LMI_X3
LMI_X3_opt = Xopt - Zopt;

fprintf('\n - LMI Check: All results below should be >= 0.\n');
[[KYP_LMI_opt{:,2}],min(real(eig(KYP_R_opt))),min(real(eig(LMI_X3_opt)))]
% fprintf(['\n - The condition numbers are: ',...
%     '\n\t\t cond(KYP_LMI_opt) = %e,',...
%     '\n\t\t cond(KYP_R_opt) = %e,',...
%     '\n\t\t cond(LMI_X3_opt) = %e.\n'],...
%     cond(KYP_LMI_opt),cond(KYP_R_opt),cond(LMI_X3_opt));
end
