function [F,copt,Mmopt,Plyapopt] = robsyn_FF_lmilab(Pt,PSIICT,R,MULT,nw,nv,nu)
%%ROBFILTSYN synthesizes a robust feedforward controller.
% Input arguments:
% (1) Pt is the plant.
% (2) PSIICT are the IQC filters.
% (3) R is part of the IQC filters. See Jonsson_99TAC.
% (4) nw is number of uncertain inputs of Pt.
% (5) nv is the number of uncertain outputs of Pt.
% (6) nu is the number of measurements sent to the estimator.
% Output arguments:
% (1) F is the synthesized estimator.
% (2) copt is the L2-L2 norm of lft(DELTA,G).
% (3) Mmopt is the IQC multiplier in the time domain.
% (4) Plyapopt is the Lyapunov function matrix.

%% Interconnection of plant and IQC filters
PICT = connect(Pt,PSIICT,Pt.InputName,...
    [PSIICT.OutputName;Pt.OutputName(nv+1:end)]);

%% Extract SS matrices
A = PICT.a; n = size(A,1); % state dimension
B1 = PICT(:,1:nw).b; % uncertain inputs
B2 = PICT(:,nw+1:end).b; m = size(B2,2); % disturbance inputs
nzPsi = size(PSIICT,1); % output size of IQC filters
C1 = PICT(1:nzPsi,:).c; % filtered uncertain outputs
C2 = PICT(nzPsi+1:nzPsi+nu,:).c; % measurements sent to estimator
p = nu; % size of measurements sent to estimator
C3 = -1.*(PICT(nzPsi+nu+1:end,:).c); % outputs to be estimated
nz = size(C3,1); % size of outputs to be estimated
D11 = PICT(1:nzPsi,1:nw).d;
D12 = PICT(1:nzPsi,nw+1:end).d;
D21 = PICT(nzPsi+1:nzPsi+nu,1:nw).d;
D22 = PICT(nzPsi+1:nzPsi+nu,nw+1:end).d;
D31 = -1.*(PICT(nzPsi+nu+1:end,1:nw).d);
D32 = -1.*(PICT(nzPsi+nu+1:end,nw+1:end).d);
switch MULT
    case {'harmonic'}
        [AR,BR,CR,DR] = ssdata(R); nstate_R = size(AR,1); ninput_R = size(DR,2);
end
nPt = size(Pt.a,1); % state dimension of Pt

%% LMILab implementation
% Set Uflag = 1 to enforce U>=0 but not KYP_R>=0
% Set Uflag = 0 to enforce KYP_R>=0 but not U>=0
Uflag = 1;

% Set CouplingFlag = 1 to enforce coupling between LMIs.
% Set CouplingFlag = 0 to not enforce this coupling.
% NOTE: If CouplingFlag == 1, then Uflag MUST be set equal to zero for it
% to have any impact. See if/else statements below.
CouplingFlag = 0;

% Initialize
setlmis([]);

% Definition of variables in LMI
X = lmivar(1,[n,1]); % (n,n) symmetric
Yhat = lmivar(1,[n,1]); % (n,n) symmetric
L = lmivar(2,[n,n]); % (n,n) full block
Bbar = lmivar(2,[n,p]); % (n,p) full block
Cbar = lmivar(2,[nz,n]); % (nz,n) full block
DF = lmivar(2,[nz,p]); % (nz,p) full block

switch MULT
    case {'harmonic'}
        szU = nzPsi/3; % size of U
        [U,~,Udec] = lmivar(1,[szU,1]); % (szU,szU) symmetric
        Mm = lmivar(3,blkdiag(Udec,Udec,-Udec)); % as given in Jonsson_99TAC
    case {'zamesfalb','sector'}
        [lm,~,lmdec] = lmivar(1,[1 1]);
        Mm = lmivar(3,lmdec*blkdiag(eye(nv),-eye(nw)));
end

[gam,ngam,gamdec] = lmivar(1,[1,1]); % (1,1) symmetric
gam_Im = lmivar(3,gamdec*eye(m));
gam_Inz = lmivar(3,gamdec*eye(nz));
switch MULT
    case {'harmonic'}
        if nstate_R>0 && ~Uflag
            Y = lmivar(1,[nstate_R,1]); % (nstate_R,nstate_R) symmetric
        end
end

% Define LMI #1: KYP_LMI < 0
KYP_LMI = newlmi;
lmiterm([+KYP_LMI 1 1 Yhat],A',eye(n),'s'); % A'*Yhat + Yhat*A
lmiterm([+KYP_LMI 1 2 X],A',eye(n)); % A'*X
lmiterm([+KYP_LMI 1 2 Yhat],eye(n),A); % Yhat*A
lmiterm([+KYP_LMI 1 3 Yhat],eye(n),B1); % Yhat*B1
lmiterm([+KYP_LMI 1 4 Yhat],eye(n),B2); % Yhat*B2
lmiterm([+KYP_LMI 1 5 0],-C3'); % -C3'
lmiterm([+KYP_LMI 2 2 X],A',eye(n),'s'); % A'*X + X*A
lmiterm([+KYP_LMI 2 3 X],eye(n),B1); % X*B1
lmiterm([+KYP_LMI 2 4 X],eye(n),B2); % X*B2
lmiterm([+KYP_LMI 2 5 0],-C3'); % -C3'
lmiterm([+KYP_LMI 3 5 0],-D31'); % -D31'
lmiterm([-KYP_LMI 4 4 gam_Im],eye(m),eye(m)); % -gm.*eye(m)
lmiterm([+KYP_LMI 4 5 0],-D32'); % -D32'
lmiterm([-KYP_LMI 5 5 gam_Inz],eye(nz),eye(nz)); % -gm.*eye(nz)

lmiterm([+KYP_LMI 1 1 Mm],C1',C1); % C1'*Mm*C1
lmiterm([+KYP_LMI 1 2 Mm],C1',C1); % C1'*Mm*C1
lmiterm([+KYP_LMI 1 3 Mm],C1',D11); % C1'*Mm*D11
lmiterm([+KYP_LMI 1 4 Mm],C1',D12); % C1'*Mm*D12
lmiterm([+KYP_LMI 2 2 Mm],C1',C1); % C1'*Mm*C1
lmiterm([+KYP_LMI 2 3 Mm],C1',D11); % C1'*Mm*D11
lmiterm([+KYP_LMI 2 4 Mm],C1',D12); % C1'*Mm*D12
lmiterm([+KYP_LMI 3 3 Mm],D11',D11); % D11'*Mm*D11
lmiterm([+KYP_LMI 3 4 Mm],D11',D12); % D11'*Mm*D12
lmiterm([+KYP_LMI 4 4 Mm],D12',D12); % D12'*Mm*D12

lmiterm([+KYP_LMI 1 2 -L],1,1) % L'
lmiterm([+KYP_LMI 1 5 -DF],C2',1); % C2'*DF'
lmiterm([+KYP_LMI 1 5 -Cbar],1,1); % Cbar'
lmiterm([+KYP_LMI 2 2 Bbar],1,C2,'s'); % Bbar*C2 + C2'*Bbar'
lmiterm([+KYP_LMI 2 3 Bbar],1,D21); % Bbar*D21
lmiterm([+KYP_LMI 2 4 Bbar],1,D22); % Bbar*D22
lmiterm([+KYP_LMI 2 5 -DF],C2',1); % C2'*DF'
lmiterm([+KYP_LMI 3 5 -DF],D21',1); % D21'*DF'
lmiterm([+KYP_LMI 4 5 -DF],D22',1); % D22'*DF'

% Define LMI #2: 0 < KYP_R
% Enforce positive definiteness of R(jw)'*U*R(jw) on jw-axis.
KYP_R = newlmi;
switch MULT
    case {'harmonic'}
        if nstate_R>0 && ~Uflag
            lmiterm([-KYP_R 1 1 Y],AR',eye(nstate_R),'s'); % AR'*Y + Y*AR
            lmiterm([-KYP_R 1 2 Y],eye(nstate_R),BR); % Y*BR
            
            lmiterm([-KYP_R 1 1 U],CR',CR); % CR'*U*CR
            lmiterm([-KYP_R 1 2 U],CR',DR); % CR'*U*DR
            lmiterm([-KYP_R 2 2 U],DR',DR); % DR'*U*DR
        else
            lmiterm([-KYP_R 1 1 U],1,1); % U
        end
    case {'zamesfalb','sector'}
        lmiterm([-KYP_R 1 1 lm],1,1); % enforce lm>=0
end

% Define LMI #3: Yhat < X (X3 > 0 ensures a stable filter is synthesized)
LMI_X3 = newlmi;
lmiterm([+LMI_X3 1 1 Yhat],1,1); % Yhat
lmiterm([-LMI_X3 1 1 X],1,1); % X

% Enforce coupling between Yhat and Y to enforce nominal stability
% See Scherer_08AUT, Theorem 4, equation 28
switch MULT
    case {'harmonic'}
        if nstate_R>0 && ~Uflag && CouplingFlag
            T_coup = [eye(nPt),zeros(nPt,nstate_R);...
                zeros(2*nstate_R,nPt),zeros(2*nstate_R,nstate_R);...
                zeros(nstate_R,nPt),eye(nstate_R)];
            %LMI_coup = T_coup'*Yhat*T_coup - ...
            %   [zeros(nPt,nPt),zeros(nPt,nstate_R);...
            %   zeros(nstate_R,nPt),Y];
            
            % Define LMI #4: coupling LMI
            LMI_coup = newlmi;
            lmiterm([+LMI_coup 1 1 0],0);
            lmiterm([+LMI_coup 2 2 Y],1,1); % Y
            %     lmiterm([-LMI_coup 0 0 0],T_coup);
            lmiterm([-LMI_coup 1 1 Yhat],T_coup',T_coup); % Yhat
        end
end

% Gain minimization subject to LMI constraints: mincx solver
lmisys = getlmis;
ndec = decnbr(lmisys); % total number of decision variables
c = zeros(ndec,1); % weighting vector for decision variables
c(ngam) = 1; % pick gamma as the cost function
% lmiopt = [];             % Default options (Display on)
lmiopt = [0 0 0 0 1];   % Turn off LMILab Display
% lmiopt = [1e-4 400 100 100 1];
% lmiopt = [1e-4 250 -1 50 0]; % Custom options (Display on)
[copt,xopt] = mincx(lmisys,c,lmiopt);

%% Verify LMI feasibility through reconstruction
Xopt = dec2mat(lmisys,xopt,X);
Yhatopt = dec2mat(lmisys,xopt,Yhat);
Lopt = dec2mat(lmisys,xopt,L);
Bbaropt = dec2mat(lmisys,xopt,Bbar);
Cbaropt = dec2mat(lmisys,xopt,Cbar);
DFopt = dec2mat(lmisys,xopt,DF);

switch MULT
    case {'harmonic'}
        Uopt = dec2mat(lmisys,xopt,U);
        Mmopt = blkdiag(Uopt,Uopt,-Uopt); % as given in Jonsson_99TAC
    case {'zamesfalb','sector'}
        lmopt = dec2mat(lmisys,xopt,lm);
        Mmopt = lmopt*blkdiag(eye(nv),-eye(nw));
end

gmopt = dec2mat(lmisys,xopt,gam);

switch MULT
    case {'harmonic'}
        if nstate_R > 0 && ~Uflag
            Yopt = dec2mat(lmisys,xopt,Y);
        end
end

% Reconstruct KYP_LMI
THp1T = ...
    [A'*Yhatopt + Yhatopt*A, A'*Xopt + Yhatopt*A, Yhatopt*B1, Yhatopt*B2, -C3';...
    A'*Yhatopt + Xopt*A, A'*Xopt + Xopt*A, Xopt*B1, Xopt*B2, -C3';...
    B1'*Yhatopt, B1'*Xopt, zeros(nw,nw), zeros(nw,m), -D31';...
    B2'*Yhatopt, B2'*Xopt, zeros(m,nw), -gmopt.*eye(m), -D32';...
    -C3, -C3, -D31, -D32, -gmopt.*eye(nz)];

THp2T = [([C1, C1, D11, D12]'*...
    Mmopt*...
    [C1, C1, D11, D12]),...
    zeros(2*n+nw+m,nz);...
    zeros(nz,2*n+nw+m), zeros(nz,nz)];

TQFRT = ...
    [zeros(n,2*n+nw+m+nz);...
    Lopt, Bbaropt*C2, Bbaropt*D21, Bbaropt*D22, zeros(n,nz);...
    zeros(nw,2*n+nw+m+nz);...
    zeros(m,2*n+nw+m+nz);...
    DFopt*C2 + Cbaropt, DFopt*C2, DFopt*D21, DFopt*D22, zeros(nz,nz)];

TRFQT = TQFRT';

KYP_LMI_opt = THp1T + THp2T + TQFRT + TRFQT;

% Reconstruct KYP_R
switch MULT
    case {'harmonic'}
        if nstate_R > 0 && ~Uflag
            KYP_R_opt = [AR'*Yopt + Yopt*AR,Yopt*BR;...
                BR'*Yopt,zeros(ninput_R)] + ...
                [CR,DR]'*Uopt*[CR,DR];
        else
            KYP_R_opt = Uopt;
        end
    case {'zamesfalb','sector'}
        KYP_R_opt = lmopt;
end

% Reconstruct LMI_X3
LMI_X3_opt = Xopt - Yhatopt;

fprintf('\n - LMI Check: All results below should be >= 0.\n');
[-max(real(eig(KYP_LMI_opt))),min(real(eig(KYP_R_opt))),min(real(eig(LMI_X3_opt)))]
% fprintf(['\n - The condition numbers are: ',...
%     '\n\t\t cond(KYP_LMI_opt) = %e,',...
%     '\n\t\t cond(KYP_R_opt) = %e,',...
%     '\n\t\t cond(LMI_X3_opt) = %e.\n'],...
%     cond(KYP_LMI_opt),cond(KYP_R_opt),cond(LMI_X3_opt));

%% Lyapunov matrix reconstruction
% SCHERER: LRM = 1 for X2 = eye(n); X3 = inv(X - Yhat);
% GAHINET: LRM = 0 for X2 = chol(X - Yhat, 'lower'); X3 = eye(n);
LRM = 0;

if LRM
    X2opt = eye(n);
    X3opt = inv(Xopt - Yhatopt);
else
    X2opt = chol(Xopt - Yhatopt,'lower');
    X3opt = eye(n);
end
Plyapopt = [Xopt, X2opt; X2opt', X3opt];

%% Filter reconstruction
% Filter state-space matrices
CF = -Cbaropt*inv(X2opt')*X3opt;
BF = inv(X2opt)*Bbaropt;
AF = (BF*C2 - inv(X2opt)*Lopt)*inv(X2opt')*X3opt;
% DF is directly synthesized
F = ss(AF,BF,CF,DFopt);

%% Input/output grouping for filter
% state names of filter
FStateName = cell(1,n);
for k = 1:n
    FStateName{1,k} = ['xf(',num2str(k),')'];
end

% input (y) to filter
if length(F.InputName) == 1
    FInpName = 'y(1)';
else
    FInpName = 'y'; % automatic vectorization expected
end

% output (xhat) from filter
if length(F.OutputName) == 1
    FOutName = 'zhat(1)';
else
    FOutName = 'zhat'; % automatic vectorization expected
end

% Set properties for F
set(F,'StateName',FStateName,...
    'InputName',FInpName,'OutputName',FOutName,...
    'InputGroup',struct('Measurement',1:length(F.InputName)),...
    'OutputGroup',struct('OutputEstimate',1:length(F.OutputName)));

%% Print results
fprintf('\n - [SYN:LMILab] The L2-L2 norm of lft(DELTA,Gt) is copt = %4.2f.\n',copt);
fprintf('\n - The condition number of Plyapopt is %e.\n',cond(Plyapopt));
fprintf(['\n - Stability of synthesized filter is ',...
    '(1-stable,0-unstable): %d.\n'],isstable(F));

end
