function [copt,Mmopt,Plyapopt] = wcgain_FFLPV_lmilab(G,PSIICT,R,MULT,nw,nv)
%%WCGANALYSIS returns the worst-case gain using IQC-based stability analysis.
% Input arguments:
% (1) G is the plant.
% (2) PSIICT are the IQC filters.
% (3) R is part of the IQC filters. See Jonsson_99TAC.
% (4) nw is number of uncertain inputs of G.
% (5) nv is the number of uncertain outputs of G.
% Output arguments:
% (1) copt is the L2-L2 norm of lft(DELTA,G).
% (2) Mmopt is the IQC multiplier in the time domain.
% (3) Plyapopt is the Lyapunov function matrix.

%% Interconnection of plant and IQC filters
GICT = connect(G,PSIICT,G.InputName,...
    [PSIICT.OutputName;G.OutputName(nv+1:end)]);
[nr,nc] = size(GICT);
% nr and nc are used in next section because 'end' does not work with PSS

%% Extract SS matrices
[A,B,C,D] = ssdata(GICT);

n = size(A,1); % state dimension
B1 = B(:,1:nw); % uncertain inputs
B2 = B(:,nw+1:nc); m = size(B2,2); % disturbance inputs
nzPsi = size(PSIICT,1); % output size of IQC filters
C1 = C(1:nzPsi,:); % filtered uncertain outputs
C2 = C(nzPsi+1:nr,:); p = size(C2,1); % performance outputs
D11 = D(1:nzPsi,1:nw);
D12 = D(1:nzPsi,nw+1:nc);
D21 = D(nzPsi+1:nr,1:nw);
D22 = D(nzPsi+1:nr,nw+1:nc);

% A = GICT.a; n = size(A,1); % state dimension
% B1 = GICT(:,1:nw).b; % uncertain inputs
% B2 = GICT(:,nw+1:nc).b; m = size(B2,2); % disturbance inputs
% nzPsi = size(PSIICT,1); % output size of IQC filters
% C1 = GICT(1:nzPsi,:).c; % filtered uncertain outputs
% C2 = GICT(nzPsi+1:nr,:).c; p = size(C2,1); % performance outputs
% D11 = GICT(1:nzPsi,1:nw).d;
% D12 = GICT(1:nzPsi,nw+1:nc).d;
% D21 = GICT(nzPsi+1:nr,1:nw).d;
% D22 = GICT(nzPsi+1:nr,nw+1:nc).d;
switch MULT
    case {'harmonic'}
        [AR,BR,CR,DR] = ssdata(R); nstate_R = size(AR,1); ninput_R = size(DR,2);
end
% Number of grid points in PSS object (currently only supports 1D grid)
ng = size(GICT.Parameter.rho.GridData,1); % number of grid points

%% LMILab implementation
% Set Uflag = 1 to enforce U>=0 but not KYP_R>=0
% Set Uflag = 0 to enforce KYP_R>=0 but not U>=0
Uflag = 1;

% Initialize
setlmis([]);

% Define Vars
Plyap = lmivar(1,[n 1]);
switch MULT
    case {'harmonic'}
        szU = nzPsi/3;
        [U,~,Udec] = lmivar(1,[szU 1]);
        Mm = lmivar(3,blkdiag(Udec,Udec,-Udec));
    case {'zamesfalb','sector'}
        [lm,~,lmdec] = lmivar(1,[1 1]);
        Mm = lmivar(3,lmdec*blkdiag(eye(nv),-eye(nw)));
end
[gam,ngam,gamdec] = lmivar(1,[1 1]);
gam_Im = lmivar(3,gamdec*eye(m));
gam_Ip = lmivar(3,gamdec*eye(p));
switch MULT
    case {'harmonic'}
        if nstate_R>0 && ~Uflag
            Y = lmivar(1,[nstate_R,1]); % (nstate_R,nstate_R) symmetric
        end
end

% Define LMIs at each grid point
LMItag = zeros(ng,1); % ng grid points
for kk = 1:ng
    % Extract data at a grid point
    Ap = A.Data(:,:,kk);
    B1p = B1.Data(:,:,kk);
    B2p = B2.Data(:,:,kk);
    C1p = C1.Data(:,:,kk);
    C2p = C2.Data(:,:,kk);
    D11p = D11.Data(:,:,kk);
    D12p = D12.Data(:,:,kk);
    D21p = D21.Data(:,:,kk);
    D22p = D22.Data(:,:,kk);
    
    % Define LMI #1
    LMItag(kk,1) = newlmi;
    lmiterm([LMItag(kk,1) 1 1 Plyap],Ap',eye(n),'s');
    lmiterm([LMItag(kk,1) 1 2 Plyap],eye(n),B1p);
    lmiterm([LMItag(kk,1) 1 3 Plyap],eye(n),B2p);
    lmiterm([LMItag(kk,1) 1 4 0],C2p');
    lmiterm([LMItag(kk,1) 2 4 0],D21p');
    lmiterm([-LMItag(kk,1) 3 3 gam_Im],eye(m),eye(m));
    lmiterm([LMItag(kk,1) 3 4 0],D22p');
    lmiterm([-LMItag(kk,1) 4 4 gam_Ip],eye(p),eye(p));
    
    lmiterm([LMItag(kk,1) 1 1 Mm],C1p',C1p);
    lmiterm([LMItag(kk,1) 1 2 Mm],C1p',D11p);
    lmiterm([LMItag(kk,1) 1 3 Mm],C1p',D12p);
    lmiterm([LMItag(kk,1) 2 2 Mm],D11p',D11p);
    lmiterm([LMItag(kk,1) 2 3 Mm],D11p',D12p);
    lmiterm([LMItag(kk,1) 3 3 Mm],D12p',D12p);
end
    
% Define LMI #2
KYP_R = newlmi;
switch MULT
    case {'harmonic'}
        if nstate_R>0 && ~Uflag % enforce KYP_R>=0
            lmiterm([-KYP_R 1 1 Y],AR',eye(nstate_R),'s');
            lmiterm([-KYP_R 1 2 Y],eye(nstate_R),BR);
            
            lmiterm([-KYP_R 1 1 U],CR',CR);
            lmiterm([-KYP_R 1 2 U],CR',DR);
            lmiterm([-KYP_R 2 2 U],DR',DR);
        else
            lmiterm([-KYP_R 1 1 U],eye(szU),eye(szU)); % enforce U>=0
        end
    case {'zamesfalb','sector'}
        lmiterm([-KYP_R 1 1 lm],1,1); % enforce lm>=0
end

% Solve lmi
lmisys = getlmis;
ndec = decnbr(lmisys); % total number of decision variables
c = zeros(ndec,1);
c(ngam)=1;
% lmiopt = [];             % Default options (Display on)
lmiopt = [0 0 0 0 1];   % Turn off LMILab Display
% lmiopt = [1e-4 200 108 100 0];
[copt,xopt] = mincx(lmisys,c,lmiopt);
fprintf('\n - [ANA:LMILab] Upper-bound on L2-L2 norm of lft(DELTA,G) is %4.2f.\n',copt);

%% Verify LMI Solution
% Compute optimal values of decision variables
Plyapopt = dec2mat(lmisys,xopt,Plyap);
switch MULT
    case {'harmonic'}
        Uopt = dec2mat(lmisys,xopt,U);
        Mmopt = blkdiag(Uopt,Uopt,-Uopt); % as given in Jonsson_99TAC
    case {'zamesfalb','sector'}
        lmopt = dec2mat(lmisys,xopt,lm);
        Mmopt = lmopt*blkdiag(eye(nv),-eye(nw));
end
gopt = dec2mat(lmisys,xopt,gam);
switch MULT
    case {'harmonic'}
        if nstate_R > 0 && ~Uflag
            Yopt = dec2mat(lmisys,xopt,Y);
        end
end

% Reconstruct LMIs at each grid point
KYP_LMI_opt = cell(ng,2); % ng grid points, second column stores eigenvalue
for kk = 1:ng
    % Extract data at a grid point
    Ap = A.Data(:,:,kk);
    B1p = B1.Data(:,:,kk);
    B2p = B2.Data(:,:,kk);
    C1p = C1.Data(:,:,kk);
    C2p = C2.Data(:,:,kk);
    D11p = D11.Data(:,:,kk);
    D12p = D12.Data(:,:,kk);
    D21p = D21.Data(:,:,kk);
    D22p = D22.Data(:,:,kk);
    
    % Reconstruct LMI #1
    T1 = ...
        [Ap'*Plyapopt + Plyapopt*Ap, Plyapopt*B1p, Plyapopt*B2p, C2p';...
        B1p'*Plyapopt, zeros(nw,nw), zeros(nw,m), D21p';...
        B2p'*Plyapopt, zeros(m,nw), -gopt.*eye(m), D22p';...
        C2p, D21p, D22p, -gopt.*eye(p)];
    
    T2 = [([C1p, D11p, D12p]'*...
        Mmopt*...
        [C1p, D11p, D12p]),...
        zeros(n+nw+m,p);...
        zeros(p,n+nw+m), zeros(p,p)];
    KYP_LMI_opt{kk,1} = T1 + T2;
    KYP_LMI_opt{kk,2} = -max(real(eig(KYP_LMI_opt{kk,1})));
end

% Reconstruct LMI #2
switch MULT
    case {'harmonic'}
        if nstate_R > 0 && ~Uflag
            KYP_R_opt = [AR'*Yopt + Yopt*AR,Yopt*BR;...
                BR'*Yopt,zeros(ninput_R)] + ...
                [CR,DR]'*Uopt*[CR,DR];
        else
            KYP_R_opt = Uopt;
        end
    case {'zamesfalb','sector'}
        KYP_R_opt = lmopt;
end

fprintf('\n - LMI Check: All results should be >=0.\n');
[[KYP_LMI_opt{:,2}],min(real(eig(KYP_R_opt)))]

end
