function [copt,Mmopt,Plyapopt] = wcgain_FF_lmilab(G,PSIICT,R,MULT,nw,nv)
%%WCGANALYSIS returns the worst-case gain using IQC-based stability analysis.
% Input arguments:
% (1) G is the plant.
% (2) PSIICT are the IQC filters.
% (3) R is part of the IQC filters. See Jonsson_99TAC.
% (4) nw is number of uncertain inputs of G.
% (5) nv is the number of uncertain outputs of G.
% Output arguments:
% (1) copt is the L2-L2 norm of lft(DELTA,G).
% (2) Mmopt is the IQC multiplier in the time domain.
% (3) Plyapopt is the Lyapunov function matrix.

%% Interconnection of plant and IQC filters
GICT = connect(G,PSIICT,G.InputName,...
    [PSIICT.OutputName;G.OutputName(nv+1:end)]);

%% Extract SS matrices
A = GICT.a; n = size(A,1); % state dimension
B1 = GICT(:,1:nw).b; % uncertain inputs
B2 = GICT(:,nw+1:end).b; m = size(B2,2); % disturbance inputs
nzPsi = size(PSIICT,1); % output size of IQC filters
C1 = GICT(1:nzPsi,:).c; % filtered uncertain outputs
C2 = GICT(nzPsi+1:end,:).c; p = size(C2,1); % performance outputs
D11 = GICT(1:nzPsi,1:nw).d;
D12 = GICT(1:nzPsi,nw+1:end).d;
D21 = GICT(nzPsi+1:end,1:nw).d;
D22 = GICT(nzPsi+1:end,nw+1:end).d;
switch MULT
    case {'harmonic'}
        [AR,BR,CR,DR] = ssdata(R); nstate_R = size(AR,1); ninput_R = size(DR,2);
end

%% LMILab implementation
% Set Uflag = 1 to enforce U>=0 but not KYP_R>=0
% Set Uflag = 0 to enforce KYP_R>=0 but not U>=0
Uflag = 1;

% Initialize
setlmis([]);

% Define Vars
Plyap = lmivar(1,[n 1]);

switch MULT
    case {'harmonic'}
        szU = nzPsi/3;
        [U,~,Udec] = lmivar(1,[szU 1]);
        Mm = lmivar(3,blkdiag(Udec,Udec,-Udec));
    case {'zamesfalb','sector'}
        [lm,~,lmdec] = lmivar(1,[1 1]);
        Mm = lmivar(3,lmdec*blkdiag(eye(nv),-eye(nw)));
end

[gam,ngam,gamdec] = lmivar(1,[1 1]);
gam_Im = lmivar(3,gamdec*eye(m));
gam_Ip = lmivar(3,gamdec*eye(p));
switch MULT
    case {'harmonic'}
        if nstate_R>0 && ~Uflag
            Y = lmivar(1,[nstate_R,1]); % (nstate_R,nstate_R) symmetric
        end
end

% Define LMI #1
lmiterm([1 1 1 Plyap],A',eye(n),'s');
lmiterm([1 1 2 Plyap],eye(n),B1);
lmiterm([1 1 3 Plyap],eye(n),B2);
lmiterm([1 1 4 0],C2');
lmiterm([1 2 4 0],D21');
lmiterm([-1 3 3 gam_Im],eye(m),eye(m));
lmiterm([1 3 4 0],D22');
lmiterm([-1 4 4 gam_Ip],eye(p),eye(p));

lmiterm([1 1 1 Mm],C1',C1);
lmiterm([1 1 2 Mm],C1',D11);
lmiterm([1 1 3 Mm],C1',D12);
lmiterm([1 2 2 Mm],D11',D11);
lmiterm([1 2 3 Mm],D11',D12);
lmiterm([1 3 3 Mm],D12',D12);

% Define LMI #2
switch MULT
    case {'harmonic'}
        if nstate_R>0 && ~Uflag % enforce KYP_R>=0
            lmiterm([-2 1 1 Y],AR',eye(nstate_R),'s');
            lmiterm([-2 1 2 Y],eye(nstate_R),BR);
            
            lmiterm([-2 1 1 U],CR',CR);
            lmiterm([-2 1 2 U],CR',DR);
            lmiterm([-2 2 2 U],DR',DR);
        else
            lmiterm([-2 1 1 U],eye(szU),eye(szU)); % enforce U>=0
        end
    case {'zamesfalb','sector'}
        lmiterm([-2 1 1 lm],1,1); % enforce lm>=0
end

% Solve lmi
lmisys = getlmis;
ndec = decnbr(lmisys); % total number of decision variables
c = zeros(ndec,1);
c(ngam)=1;
% lmiopt = [];             % Default options (Display on)
lmiopt = [0 0 0 0 1];   % Turn off LMILab Display
% lmiopt = [1e-4 200 108 100 0];
[copt,xopt] = mincx(lmisys,c,lmiopt);
fprintf('\n - [ANA:LMILab] Upper-bound on L2-L2 norm of lft(DELTA,G) is %4.2f.\n',copt);

%% Verify LMI Solution
Plyapopt = dec2mat(lmisys,xopt,Plyap);

switch MULT
    case {'harmonic'}
        Uopt = dec2mat(lmisys,xopt,U);
        Mmopt = blkdiag(Uopt,Uopt,-Uopt); % as given in Jonsson_99TAC
    case {'zamesfalb','sector'}
        lmopt = dec2mat(lmisys,xopt,lm);
        Mmopt = lmopt*blkdiag(eye(nv),-eye(nw));
end

gopt = dec2mat(lmisys,xopt,gam);

switch MULT
    case {'harmonic'}
        if nstate_R > 0 && ~Uflag
            Yopt = dec2mat(lmisys,xopt,Y);
        end
end

T1 = ...
    [A'*Plyapopt + Plyapopt*A, Plyapopt*B1, Plyapopt*B2, C2';...
    B1'*Plyapopt, zeros(nw,nw), zeros(nw,m), D21';...
    B2'*Plyapopt, zeros(m,nw), -gopt.*eye(m), D22';...
    C2, D21, D22, -gopt.*eye(p)];

T2 = [([C1, D11, D12]'*...
    Mmopt*...
    [C1, D11, D12]),...
    zeros(n+nw+m,p);...
    zeros(p,n+nw+m), zeros(p,p)];
KYP_LMI_opt = T1+T2;

switch MULT
    case {'harmonic'}
        if nstate_R > 0 && ~Uflag
            KYP_R_opt = [AR'*Yopt + Yopt*AR,Yopt*BR;...
                BR'*Yopt,zeros(ninput_R)] + ...
                [CR,DR]'*Uopt*[CR,DR];
        else
            KYP_R_opt = Uopt;
        end
    case {'zamesfalb','sector'}
        KYP_R_opt = lmopt;
end

fprintf('\n - LMI Check: All results should be >=0.\n');
[-max(real(eig(KYP_LMI_opt))),min(real(eig(KYP_R_opt)))]

end
